from __future__ import annotations

import enum
import json
import os
from collections.abc import Mapping
from contextlib import asynccontextmanager
from typing import Any, AsyncIterator, Dict, List, NotRequired, Optional, TypedDict, cast

import aiohttp


class QuotaTypes(enum.StrEnum):
    DIRECTORY = "directory"
    USER = "user"
    GROUP = "group"


class QuotaThresholds(TypedDict):
    hard: NotRequired[int]
    soft: NotRequired[int]


class OneFSClient:
    endpoint: str
    user: str
    password: str
    api_version: str
    system_name: Optional[str]
    _session: aiohttp.ClientSession

    def __init__(
        self,
        endpoint: str,
        user: str,
        password: str,
        *,
        api_version: str = "12",
        system_name: Optional[str] = "nfs",
    ) -> None:
        self.endpoint = endpoint
        self.user = user
        self.password = password
        self.api_version = api_version
        self.system_name = system_name
        self._session = aiohttp.ClientSession()

    async def aclose(self) -> None:
        await self._session.close()

    async def get_metadata(self) -> Mapping[str, Any]:
        cluster_metadata = await self.get_cluster_metadata()
        node_metadata = await self.get_node_metadata()
        volume_cluster = {
            "cluster": json.dumps(cluster_metadata),
            "nodes": json.dumps(node_metadata),
        }
        return volume_cluster

    @asynccontextmanager
    async def _request(
        self,
        method: str,
        path: str,
        **kwargs,
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        async with self._session.request(
            method,
            f"{self.endpoint}/platform/{self.api_version}/{path}",
            auth=aiohttp.BasicAuth(self.user, self.password),
            ssl=False,
            raise_for_status=True,
            **kwargs,
        ) as resp:
            yield resp

    async def get_usage(self) -> Mapping[str, int]:
        async with self._request("GET", "storagepool/storagepools") as resp:
            data = await resp.json()
        return {
            "capacity_bytes": int(data["storagepools"][0]["usage"]["total_bytes"]),
            "used_bytes": int(data["storagepools"][0]["usage"]["used_bytes"]),
        }

    async def get_list_lnn(self) -> List[int]:
        async with self._request("GET", "storagepool/storagepools") as resp:
            data = await resp.json()
        return data["storagepools"][0]["lnns"]

    async def get_node_hardware_info_by_lnn(self, lnn) -> Mapping[str, Any]:
        async with self._request("GET", f"cluster/nodes/{lnn}/hardware") as resp:
            data = await resp.json()
            node = data["nodes"][0]
        return {
            "id": node["id"],
            "model": node["product"],
            "configuration": node["configuration_id"],
            "serial_number": node["serial_number"],
        }

    async def get_node_status_by_lnn(self, lnn) -> Mapping[str, Any]:
        async with self._request("GET", f"cluster/nodes/{lnn}/status/nvram") as resp:
            data = await resp.json()
            node = data["nodes"][0]
        return {
            "batteries": node["batteries"],
            # "capacity": node["capacity"],
        }

    async def get_cluster_metadata(self) -> List[Dict[str, Any]]:
        try:
            cluster_metadata = []
            cluster_metadata.append({
                "config": await self.get_cluster_config(),
                "interface": await self.get_cluster_interface(),
            })
            return cluster_metadata
        except Exception as e:
            raise (e)

    async def get_cluster_config(self) -> Mapping[str, Any]:
        async with self._request("GET", "cluster/config") as resp:
            data = await resp.json()
        return {
            "name": data["name"],
            "release": data["onefs_version"]["release"],
            "build": data["onefs_version"]["build"],
            "local_lnn": data["local_lnn"],
        }

    async def get_cluster_interface(self) -> List[Mapping[str, Any]]:
        async with self._request("GET", "network/interfaces") as resp:
            data = await resp.json()
        return data["interfaces"]

    async def get_node_metadata(self) -> List[Dict[str, Mapping[str, Any]]]:
        try:
            lnns = await self.get_list_lnn()
            node_metadata = []
            for lnn in lnns:
                node_metadata.append({
                    "hardware": await self.get_node_hardware_info_by_lnn(lnn),
                    "status": await self.get_node_status_by_lnn(lnn),
                })
            return node_metadata
        except Exception as e:
            raise (e)

    async def get_drive_stats(self) -> Mapping[str, Any]:
        async with self._request("GET", "statistics/summary/drive") as resp:
            data = await resp.json()
        return data["drive"]

    async def get_protocol_stats(self) -> Mapping[str, Any]:
        async with self._request("GET", "statistics/summary/protocol-stats") as resp:
            data = await resp.json()
        return data["protocol-stats"]

    async def get_workload_stats(self) -> Mapping[str, Any]:
        params = {}
        if self.system_name is not None:
            params = {"system_names": self.system_name}
        async with self._request(
            "GET",
            "statistics/summary/workload",
            params=params,
        ) as resp:
            data = await resp.json()
        workload_stats = cast(list[Mapping[str, Any]], data["workload"])
        return workload_stats[0]

    async def get_system_stats(self) -> Mapping[str, Any]:
        async with self._request("GET", "statistics/summary/system") as resp:
            data = await resp.json()
        system_stats = cast(list[Mapping[str, Any]], data["system"])
        return system_stats[0]

    async def list_all_quota(self) -> Mapping[str, Any]:
        async with self._request("GET", "quota/quotas") as resp:
            data = await resp.json()
        return data["quotas"]

    async def get_quota(self, quota_id: str) -> Mapping[str, Any]:
        async with self._request("GET", f"quota/quotas/{quota_id}") as resp:
            data = await resp.json()
        return data["quotas"][0]

    async def create_quota(
        self,
        path: os.PathLike,
        type_: QuotaTypes,
        thresholds: QuotaThresholds,
    ) -> Mapping[str, Any]:
        data = {
            "path": os.fsdecode(path),
            "type": type_.value,
            "include_snapshots": False,
            "thresholds": thresholds,
            "thresholds_on": "fslogicalsize",
            "enforced": True,
        }
        async with self._request(
            "POST",
            "quota/quotas",
            json=data,
        ) as resp:
            msg = await resp.json()
        return msg

    async def delete_quota(self, quota_id) -> None:
        async with self._request(
            "DELETE",
            f"quota/quotas/{quota_id}",
        ) as _:
            return

    async def update_quota(self, quota_id: str, thresholds: QuotaThresholds) -> None:
        data = {
            "thresholds": thresholds,
        }
        async with self._request(
            "PUT",
            f"quota/quotas/{quota_id}",
            json=data,
        ) as _:
            return



from __future__ import annotations

import enum
from datetime import datetime
from pathlib import Path, PurePath
from typing import Any, Final, Mapping, Optional

import attrs
import trafaret as t

from ai.backend.common import validators as tx
from ai.backend.common.types import QuotaConfig, VFolderID

__all__ = (
    "Sentinel",
    "SENTINEL",
    "FSPerfMetric",
    "CapacityUsage",
    "VolumeInfo",
    "VFolderID",
    "TreeUsage",
    "QuotaConfig",
    "Stat",
    "DirEntry",
    "DirEntryType",
)


class Sentinel(enum.Enum):
    token = 0


SENTINEL: Final = Sentinel.token


@attrs.define(slots=True, frozen=True)
class FSPerfMetric:
    # iops
    iops_read: int
    iops_write: int
    # thruput
    io_bytes_read: int
    io_bytes_write: int
    # latency
    io_usec_read: float
    io_usec_write: float


@attrs.define(slots=True, frozen=True)
class CapacityUsage:
    used_bytes: int
    capacity_bytes: int


@attrs.define(slots=True, frozen=True)
class QuotaUsage:
    used_bytes: int
    limit_bytes: int


@attrs.define(slots=True, frozen=True)
class VolumeInfo:
    backend: str
    path: Path
    fsprefix: Optional[PurePath]
    options: Optional[Mapping[str, Any]]

    @classmethod
    def as_trafaret(cls) -> t.Trafaret:
        return t.Dict(
            {
                t.Key("backend"): t.String,
                t.Key("path"): tx.Path(type="dir"),
                t.Key("fsprefix", default="."): tx.PurePath(relative_only=True),
                t.Key("options", default=None): t.Null | t.Mapping(t.String, t.Any),
            },
        )


@attrs.define(slots=True, frozen=True)
class TreeUsage:
    file_count: int  # TODO: refactor using DecimalSize
    used_bytes: int  # TODO: refactor using DecimalSize


@attrs.define(slots=True, frozen=True)
class Stat:
    size: int
    owner: str
    mode: int
    modified: datetime
    created: datetime


class DirEntryType(enum.Enum):
    FILE = 0
    DIRECTORY = 1
    SYMLINK = 2


@attrs.define(auto_attribs=True, slots=True, frozen=True)
class DirEntry:
    name: str
    path: Path
    type: DirEntryType
    stat: Stat
    symlink_target: str



from __future__ import annotations

import logging
from contextlib import asynccontextmanager as actxmgr
from pathlib import Path
from typing import (
    Any,
    AsyncIterator,
    Final,
    Mapping,
    Optional,
    Type,
)

import aiohttp_cors
from aiohttp import web
from aiohttp.typedefs import Middleware

from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events import (
    EventDispatcher,
    EventProducer,
)
from ai.backend.logging import BraceStyleAdapter

from .abc import AbstractVolume
from .api.client import init_client_app
from .api.manager import init_manager_app
from .cephfs import CephFSVolume
from .ddn import EXAScalerFSVolume
from .dellemc import DellEMCOneFSVolume
from .exception import InvalidVolumeError
from .gpfs import GPFSVolume
from .netapp import NetAppVolume
from .plugin import (
    BasePluginContext,
    StorageClientWebappPluginContext,
    StorageManagerWebappPluginContext,
    StoragePluginContext,
)
from .purestorage import FlashBladeVolume
from .types import VolumeInfo
from .vast import VASTVolume
from .vfs import BaseVolume
from .watcher import WatcherClient
from .weka import WekaVolume
from .xfs import XfsVolume

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

EVENT_DISPATCHER_CONSUMER_GROUP: Final = "storage-proxy"

DEFAULT_BACKENDS: Mapping[str, Type[AbstractVolume]] = {
    FlashBladeVolume.name: FlashBladeVolume,
    BaseVolume.name: BaseVolume,
    XfsVolume.name: XfsVolume,
    NetAppVolume.name: NetAppVolume,
    # NOTE: Dell EMC has two different storage: PowerStore and PowerScale (OneFS).
    #       We support the latter only for now.
    DellEMCOneFSVolume.name: DellEMCOneFSVolume,
    WekaVolume.name: WekaVolume,
    GPFSVolume.name: GPFSVolume,  # IBM SpectrumScale or GPFS
    "spectrumscale": GPFSVolume,  # IBM SpectrumScale or GPFS
    CephFSVolume.name: CephFSVolume,
    VASTVolume.name: VASTVolume,
    EXAScalerFSVolume.name: EXAScalerFSVolume,
}


async def on_prepare(request: web.Request, response: web.StreamResponse) -> None:
    response.headers["Server"] = "BackendAI"


def _init_subapp(
    pkg_name: str,
    root_app: web.Application,
    subapp: web.Application,
    global_middlewares: list[Middleware],
) -> None:
    subapp.on_response_prepare.append(on_prepare)

    async def _set_root_ctx(subapp: web.Application):
        # Allow subapp's access to the root app properties.
        # These are the public APIs exposed to plugins as well.
        subapp["ctx"] = root_app["ctx"]

    # We must copy the public interface prior to all user-defined startup signal handlers.
    subapp.on_startup.insert(0, _set_root_ctx)
    if "prefix" not in subapp:
        subapp["prefix"] = pkg_name.split(".")[-1].replace("_", "-")
    prefix = subapp["prefix"]
    root_app.add_subapp("/" + prefix, subapp)
    root_app.middlewares.extend(global_middlewares)


class RootContext:
    volumes: dict[str, AbstractVolume]
    pid: int
    etcd: AsyncEtcd
    local_config: Mapping[str, Any]
    dsn: str | None
    event_producer: EventProducer
    event_dispatcher: EventDispatcher
    watcher: WatcherClient | None

    def __init__(
        self,
        pid: int,
        pidx: int,
        node_id: str,
        local_config: Mapping[str, Any],
        etcd: AsyncEtcd,
        *,
        event_producer: EventProducer,
        event_dispatcher: EventDispatcher,
        watcher: WatcherClient | None,
        dsn: Optional[str] = None,
    ) -> None:
        self.volumes = {}
        self.pid = pid
        self.pidx = pidx
        self.node_id = node_id
        self.etcd = etcd
        self.local_config = local_config
        self.dsn = dsn
        self.event_producer = event_producer
        self.event_dispatcher = event_dispatcher
        self.watcher = watcher
        self.cors_options = {
            "*": aiohttp_cors.ResourceOptions(
                allow_credentials=False, expose_headers="*", allow_headers="*"
            ),
        }

    async def __aenter__(self) -> None:
        self.client_api_app = await init_client_app(self)
        self.manager_api_app = await init_manager_app(self)
        self.backends = {
            **DEFAULT_BACKENDS,
        }
        await self.init_storage_plugin()
        self.manager_webapp_plugin_ctx = await self.init_storage_webapp_plugin(
            StorageManagerWebappPluginContext(self.etcd, self.local_config),
            self.manager_api_app,
        )
        self.client_webapp_plugin_ctx = await self.init_storage_webapp_plugin(
            StorageClientWebappPluginContext(self.etcd, self.local_config),
            self.client_api_app,
        )

    async def init_storage_plugin(self) -> None:
        plugin_ctx = StoragePluginContext(self.etcd, self.local_config)
        await plugin_ctx.init()
        self.storage_plugin_ctx = plugin_ctx
        for plugin_name, plugin_instance in plugin_ctx.plugins.items():
            log.info("Loading storage plugin: {0}", plugin_name)
            volume_cls = plugin_instance.get_volume_class()
            self.backends[plugin_name] = volume_cls

    async def init_storage_webapp_plugin(
        self, plugin_ctx: BasePluginContext, root_app: web.Application
    ) -> BasePluginContext:
        await plugin_ctx.init()
        for plugin_name, plugin_instance in plugin_ctx.plugins.items():
            if self.pid == 0:
                log.info("Loading storage webapp plugin: {0}", plugin_name)
            subapp, global_middlewares = await plugin_instance.create_app(self.cors_options)
            _init_subapp(plugin_name, root_app, subapp, global_middlewares)
        return plugin_ctx

    def list_volumes(self) -> Mapping[str, VolumeInfo]:
        return {name: VolumeInfo(**info) for name, info in self.local_config["volume"].items()}

    async def __aexit__(self, *exc_info) -> Optional[bool]:
        for volume in self.volumes.values():
            await volume.shutdown()

        await self.storage_plugin_ctx.cleanup()
        await self.manager_webapp_plugin_ctx.cleanup()
        await self.client_webapp_plugin_ctx.cleanup()
        return None

    @actxmgr
    async def get_volume(self, name: str) -> AsyncIterator[AbstractVolume]:
        if name in self.volumes:
            yield self.volumes[name]
        else:
            try:
                volume_config = self.local_config["volume"][name]
            except KeyError:
                raise InvalidVolumeError(name)
            volume_cls: Type[AbstractVolume] = self.backends[volume_config["backend"]]
            volume_obj = volume_cls(
                local_config=self.local_config,
                mount_path=Path(volume_config["path"]),
                options=volume_config["options"] or {},
                etcd=self.etcd,
                event_dispatcher=self.event_dispatcher,
                event_producer=self.event_producer,
            )

            await volume_obj.init()
            self.volumes[name] = volume_obj

            yield volume_obj



from __future__ import annotations

import asyncio
import contextlib
import logging
import re
from typing import FrozenSet

from ai.backend.common.types import HardwareMetadata
from ai.backend.logging import BraceStyleAdapter

from ..abc import CAP_FAST_FS_SIZE, CAP_FAST_SCAN, CAP_METRIC, CAP_VFOLDER, AbstractFSOpModel
from ..types import CapacityUsage, FSPerfMetric
from ..vfs import BaseVolume
from .purity import PurityClient
from .rapidfiles import RapidFileToolsFSOpModel
from .rapidfiles_v2 import RapidFileToolsv2FSOpModel

FLASHBLADE_TOOLKIT_V2_VERSION_RE = re.compile(r"version p[a-zA-Z\d]+ \(RapidFile\) (2\..+)")
FLASHBLADE_TOOLKIT_V1_VERSION_RE = re.compile(r"p[a-zA-Z\d]+ \(RapidFile Toolkit\) (1\..+)")

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class FlashBladeVolume(BaseVolume):
    name = "purestorage"
    _toolkit_version: int | None

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._toolkit_version = None

    async def create_fsop_model(self) -> AbstractFSOpModel:
        if (await self.get_toolkit_version()) == 2:
            return RapidFileToolsv2FSOpModel(
                self.mount_path,
                self.local_config["storage-proxy"]["scandir-limit"],
            )
        else:
            return RapidFileToolsFSOpModel(
                self.mount_path,
                self.local_config["storage-proxy"]["scandir-limit"],
            )

    async def get_toolkit_version(self) -> int:
        if self._toolkit_version is not None:
            return self._toolkit_version
        try:
            proc = await asyncio.create_subprocess_exec(
                b"pdu",
                b"--version",
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.STDOUT,
            )
        except FileNotFoundError:
            self._toolkit_version = -1
            return -1
        try:
            stdout, stderr = await proc.communicate()
            if proc.returncode != 0:
                self._toolkit_version = -1
            else:
                version_line = stdout.decode().splitlines()[0]
                if FLASHBLADE_TOOLKIT_V2_VERSION_RE.match(version_line):
                    self._toolkit_version = 2
                    log.info("FlashBlade Toolkit 2 detected")
                elif FLASHBLADE_TOOLKIT_V1_VERSION_RE.match(version_line):
                    self._toolkit_version = 1
                    log.info("FlashBlade Toolkit 1 detected")
                else:
                    log.warning("Unrecogized FlashBlade Toolkit version: {}", version_line)
                    self._toolkit_version = -1
        finally:
            await proc.wait()
            assert self._toolkit_version
            return self._toolkit_version

    async def init(self) -> None:
        toolkit_version = await self.get_toolkit_version()
        if toolkit_version == -1:
            raise RuntimeError(
                "PureStorage RapidFile Toolkit is not installed. "
                "You cannot use the PureStorage backend for the stroage proxy.",
            )
        self.purity_client = PurityClient(
            self.config["purity_endpoint"],
            self.config["purity_api_token"],
            api_version=self.config["purity_api_version"],
        )
        await super().init()

    async def shutdown(self) -> None:
        await self.purity_client.aclose()

    async def get_capabilities(self) -> FrozenSet[str]:
        return frozenset(
            [
                CAP_FAST_FS_SIZE,
                CAP_VFOLDER,
                CAP_METRIC,
                CAP_FAST_SCAN,
            ],
        )

    async def get_hwinfo(self) -> HardwareMetadata:
        async with self.purity_client as client:
            metadata = await client.get_metadata()
        return {
            "status": "healthy",
            "status_info": None,
            "metadata": {
                **metadata,
            },
        }

    async def get_fs_usage(self) -> CapacityUsage:
        async with self.purity_client as client:
            usage = await client.get_usage(self.config["purity_fs_name"])
        return CapacityUsage(
            capacity_bytes=usage["capacity_bytes"],
            used_bytes=usage["used_bytes"],
        )

    async def get_performance_metric(self) -> FSPerfMetric:
        async with self.purity_client as client:
            async with contextlib.aclosing(
                client.get_nfs_metric(self.config["purity_fs_name"]),
            ) as items:
                async for item in items:
                    return FSPerfMetric(
                        iops_read=item["reads_per_sec"],
                        iops_write=item["writes_per_sec"],
                        io_bytes_read=item["read_bytes_per_sec"],
                        io_bytes_write=item["write_bytes_per_sec"],
                        io_usec_read=item["usec_per_read_op"],
                        io_usec_write=item["usec_per_write_op"],
                    )
                else:
                    raise RuntimeError(
                        "no metric found for the configured flashblade filesystem",
                    )



from __future__ import annotations

import asyncio
import errno
import functools
import logging
import os
import secrets
import shutil
import time
from collections import deque
from pathlib import Path, PurePosixPath
from typing import Any, AsyncIterator, FrozenSet, Optional, Sequence, Union, final

import aiofiles.os
import janus
import trafaret as t

from ai.backend.common.types import BinarySize, HardwareMetadata, QuotaScopeID
from ai.backend.logging import BraceStyleAdapter

from ..abc import CAP_VFOLDER, AbstractFSOpModel, AbstractQuotaModel, AbstractVolume
from ..exception import (
    ExecutionError,
    InvalidAPIParameters,
    InvalidQuotaScopeError,
    NotEmptyError,
    QuotaScopeNotFoundError,
)
from ..subproc import run
from ..types import (
    SENTINEL,
    CapacityUsage,
    DirEntry,
    DirEntryType,
    FSPerfMetric,
    QuotaConfig,
    QuotaUsage,
    Sentinel,
    Stat,
    TreeUsage,
    VFolderID,
)
from ..utils import fstime2datetime

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class BaseQuotaModel(AbstractQuotaModel):
    """
    This quota model just creates the first-level volume directories
    to split vfolders into per-user/per-project namespaces, without
    imposing any actual quota limits.
    """

    def __init__(self, mount_path: Path) -> None:
        self.mount_path = mount_path

    def mangle_qspath(self, ref: VFolderID | QuotaScopeID | str | None) -> Path:
        try:
            match ref:
                case VFolderID():
                    if ref.quota_scope_id is None:
                        return self.mount_path  # for legacy vfolder paths during migration
                    return Path(self.mount_path, ref.quota_scope_id.pathname)
                case QuotaScopeID():
                    return Path(self.mount_path, ref.pathname)
                case str():
                    typed_scope_id = QuotaScopeID.parse(ref)
                    return Path(self.mount_path, typed_scope_id.pathname)
                case None:
                    return self.mount_path  # for legacy vfolder paths during migration
                case _:
                    raise InvalidQuotaScopeError(
                        f"Invalid value format for quota scope ID: {ref!r}"
                    )
        except t.DataError:
            raise InvalidQuotaScopeError(f"Invalid value format for quota scope ID: {ref!r}")

    async def create_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        options: Optional[QuotaConfig] = None,
        extra_args: Optional[dict[str, Any]] = None,
    ) -> None:
        qspath = self.mangle_qspath(quota_scope_id)
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            lambda: qspath.mkdir(0o755, parents=True, exist_ok=False),
        )

    async def describe_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> Optional[QuotaUsage]:
        if not self.mangle_qspath(quota_scope_id).exists():
            return None

        return QuotaUsage(-1, -1)

    async def update_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        config: QuotaConfig,
    ) -> None:
        # This is a no-op.
        pass

    async def unset_quota(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> None:
        # This is a no-op.
        pass

    async def delete_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> None:
        qspath = self.mangle_qspath(quota_scope_id)
        if len([p for p in qspath.iterdir() if p.is_dir()]) > 0:
            raise NotEmptyError(quota_scope_id)
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            lambda: shutil.rmtree(qspath),
        )


class SetGIDQuotaModel(BaseQuotaModel):
    """
    This quota model uses the Linux's vanilla gid-based quota scheme
    with setgid on the first-level namespace directories for each user
    or each project.
    """

    async def create_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        options: Optional[QuotaConfig] = None,
        extra_args: Optional[dict[str, Any]] = None,
    ) -> None:
        qspath = self.mangle_qspath(quota_scope_id)
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            lambda: qspath.mkdir(0o755, parents=True, exist_ok=False),
        )
        # TODO: setgid impl.

    async def describe_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> Optional[QuotaUsage]:
        if not self.mangle_qspath(quota_scope_id).exists():
            return None
        # TODO: setgid impl.
        return QuotaUsage(-1, -1)

    async def update_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        options: QuotaConfig,
    ) -> None:
        # TODO: setgid impl.
        raise NotImplementedError

    async def unset_quota(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> None:
        # TODO: setgid impl.
        raise NotImplementedError

    async def delete_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> None:
        qspath = self.mangle_qspath(quota_scope_id)
        if len([p for p in qspath.iterdir() if p.is_dir()]) > 0:
            raise NotEmptyError(quota_scope_id)
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            lambda: shutil.rmtree(qspath),
        )
        # TODO: setgid impl.


class BaseFSOpModel(AbstractFSOpModel):
    def __init__(self, mount_path: Path, scandir_limit: int) -> None:
        self.mount_path = mount_path
        self.scandir_limit = scandir_limit

    async def copy_tree(
        self,
        src_path: Path,
        dst_path: Path,
    ) -> None:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            functools.partial(
                shutil.copytree,
                src_path,
                dst_path,
                dirs_exist_ok=True,
            ),
        )

    async def move_tree(
        self,
        src_path: Path,
        dst_path: Path,
    ) -> None:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            lambda: shutil.move(str(src_path), str(dst_path)),
        )

    async def delete_tree(
        self,
        path: Path,
    ) -> None:
        loop = asyncio.get_running_loop()
        try:
            await loop.run_in_executor(None, lambda: shutil.rmtree(path))
        except FileNotFoundError:
            pass

    def scan_tree(
        self,
        path: Path,
        *,
        recursive: bool = True,
    ) -> AsyncIterator[DirEntry]:
        q: janus.Queue[Sentinel | DirEntry] = janus.Queue()
        loop = asyncio.get_running_loop()

        def _scandir(path: Path, q: janus._SyncQueueProxy[Sentinel | DirEntry]) -> None:
            count = 0
            limit = self.scandir_limit
            next_paths: deque[Path] = deque()
            next_paths.append(path)
            while next_paths:
                next_path = next_paths.popleft()
                with os.scandir(next_path) as scanner:
                    for entry in scanner:
                        if limit > 0 and count == limit:
                            break
                        symlink_target = ""
                        entry_type = DirEntryType.FILE
                        try:
                            if entry.is_dir(follow_symlinks=False):
                                entry_type = DirEntryType.DIRECTORY
                            if entry.is_symlink():
                                entry_type = DirEntryType.SYMLINK
                                try:
                                    symlink_dst = Path(entry).resolve()
                                    symlink_dst = symlink_dst.relative_to(path)
                                except (ValueError, RuntimeError):
                                    # ValueError and ELOOP
                                    pass
                                else:
                                    symlink_target = os.fsdecode(symlink_dst)
                            entry_stat = entry.stat(follow_symlinks=False)
                        except (FileNotFoundError, PermissionError):
                            # the filesystem may be changed during scan
                            continue
                        item = DirEntry(
                            name=entry.name,
                            path=Path(entry.path).relative_to(path),
                            type=entry_type,
                            stat=Stat(
                                size=entry_stat.st_size,
                                owner=str(entry_stat.st_uid),
                                mode=entry_stat.st_mode,
                                modified=fstime2datetime(entry_stat.st_mtime),
                                created=fstime2datetime(entry_stat.st_ctime),
                            ),
                            symlink_target=symlink_target,
                        )
                        q.put(item)
                        if recursive and entry.is_dir() and not entry.is_symlink():
                            next_paths.append(Path(entry.path))
                        count += 1

        async def _scan_task(q: janus.Queue[Sentinel | DirEntry]) -> None:
            try:
                await loop.run_in_executor(None, _scandir, path, q.sync_q)
            finally:
                await q.async_q.put(SENTINEL)

        async def _aiter() -> AsyncIterator[DirEntry]:
            scan_task = asyncio.create_task(_scan_task(q))
            await asyncio.sleep(0)
            try:
                while True:
                    item = await q.async_q.get()
                    try:
                        if item is SENTINEL:
                            break
                        yield item
                    finally:
                        q.async_q.task_done()
            finally:
                await scan_task
                q.close()
                await q.wait_closed()

        return _aiter()

    async def scan_tree_usage(
        self,
        path: Path,
    ) -> TreeUsage:
        total_size = 0
        total_count = 0
        start_time = time.monotonic()
        _timeout = 30

        def _calc_usage(path: Path) -> None:
            nonlocal total_size, total_count
            next_paths: deque[Path] = deque()
            next_paths.append(path)
            while next_paths:
                next_path = next_paths.popleft()
                with os.scandir(next_path) as scanner:  # type: ignore
                    for entry in scanner:
                        try:
                            stat = entry.stat(follow_symlinks=False)
                        except (FileNotFoundError, PermissionError):
                            # the filesystem may be changed during scan
                            continue
                        total_size += stat.st_size
                        total_count += 1
                        if entry.is_dir() and not entry.is_symlink():
                            next_paths.append(Path(entry.path))
                        if total_count % 1000 == 0:
                            # Cancel if this I/O operation takes too much time.
                            if time.monotonic() - start_time > _timeout:
                                raise TimeoutError

        loop = asyncio.get_running_loop()
        try:
            await loop.run_in_executor(None, _calc_usage, path)
        except TimeoutError:
            # -1 indicates "too many"
            total_size = -1
            total_count = -1
        return TreeUsage(file_count=total_count, used_bytes=total_size)

    async def scan_tree_size(
        self,
        path: Path,
    ) -> BinarySize:
        info = await run(["du", "-hs", path])
        used_bytes, _ = info.split()
        return BinarySize.finite_from_str(used_bytes)


class BaseVolume(AbstractVolume):
    name = "vfs"

    async def create_quota_model(self) -> AbstractQuotaModel:
        return BaseQuotaModel(self.mount_path)

    async def create_fsop_model(self) -> AbstractFSOpModel:
        return BaseFSOpModel(
            self.mount_path,
            self.local_config["storage-proxy"]["scandir-limit"],
        )

    async def get_capabilities(self) -> FrozenSet[str]:
        return frozenset([CAP_VFOLDER])

    async def get_hwinfo(self) -> HardwareMetadata:
        return {
            "status": "healthy",
            "status_info": None,
            "metadata": {},
        }

    @final
    async def create_vfolder(
        self,
        vfid: VFolderID,
        exist_ok=False,
    ) -> None:
        qspath = self.quota_model.mangle_qspath(vfid)
        if not qspath.exists():
            raise QuotaScopeNotFoundError
        vfpath = self.mangle_vfpath(vfid)
        await aiofiles.os.makedirs(vfpath, 0o755, exist_ok=exist_ok)

    @final
    async def delete_vfolder(self, vfid: VFolderID) -> None:
        vfpath = self.mangle_vfpath(vfid)
        await self.fsop_model.delete_tree(vfpath)
        for p in [vfpath, vfpath.parent, vfpath.parent.parent]:
            try:
                await aiofiles.os.rmdir(p)
            except FileNotFoundError:
                pass
            except OSError as e:
                match e.errno:
                    case errno.ENOTEMPTY:
                        pass
                    case _:
                        raise

    @final
    async def clone_vfolder(
        self,
        src_vfid: VFolderID,
        dst_vfid: VFolderID,
    ) -> None:
        # check if there is enough space in the destination
        fs_usage = await self.get_fs_usage()
        vfolder_usage = await self.get_usage(src_vfid)
        if vfolder_usage.used_bytes > fs_usage.capacity_bytes - fs_usage.used_bytes:
            raise ExecutionError("Not enough space available for clone.")

        # create the target vfolder
        src_vfpath = self.mangle_vfpath(src_vfid)
        await self.create_vfolder(dst_vfid)
        dst_vfpath = self.mangle_vfpath(dst_vfid)

        # perform the file-tree copy
        try:
            await self.fsop_model.copy_tree(src_vfpath, dst_vfpath)
        except Exception:
            await self.delete_vfolder(dst_vfid)
            log.exception("clone_vfolder: error during copy_tree()")
            raise ExecutionError("Copying files from source directories failed.")

    @final
    async def get_vfolder_mount(self, vfid: VFolderID, subpath: str) -> Path:
        self.sanitize_vfpath(vfid, PurePosixPath(subpath))
        return self.mangle_vfpath(vfid).resolve()

    async def put_metadata(self, vfid: VFolderID, payload: bytes) -> None:
        vfpath = self.mangle_vfpath(vfid)
        metadata_path = vfpath / "metadata.json"
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, metadata_path.write_bytes, payload)

    async def get_metadata(self, vfid: VFolderID) -> bytes:
        vfpath = self.mangle_vfpath(vfid)
        metadata_path = vfpath / "metadata.json"
        loop = asyncio.get_running_loop()
        try:
            stat = await loop.run_in_executor(None, metadata_path.stat)
            if stat.st_size > 10 * (2**20):
                raise RuntimeError("Too large metadata (more than 10 MiB)")
            data = await loop.run_in_executor(None, metadata_path.read_bytes)
            return data
        except FileNotFoundError:
            return b""
        # Other IO errors should be bubbled up.

    async def get_performance_metric(self) -> FSPerfMetric:
        raise NotImplementedError

    async def get_fs_usage(self) -> CapacityUsage:
        loop = asyncio.get_running_loop()
        stat = await loop.run_in_executor(None, os.statvfs, self.mount_path)
        return CapacityUsage(
            capacity_bytes=BinarySize(stat.f_frsize * stat.f_blocks),
            used_bytes=BinarySize(stat.f_frsize * (stat.f_blocks - stat.f_bavail)),
        )

    @final
    async def get_usage(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath = PurePosixPath("."),
    ) -> TreeUsage:
        target_path = self.sanitize_vfpath(vfid, relpath)
        return await self.fsop_model.scan_tree_usage(target_path)

    @final
    async def get_used_bytes(self, vfid: VFolderID) -> BinarySize:
        vfpath = self.mangle_vfpath(vfid)
        return await self.fsop_model.scan_tree_size(vfpath)

    # ------ vfolder internal operations -------

    @final
    def scandir(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        *,
        recursive: bool = True,
    ) -> AsyncIterator[DirEntry]:
        target_path = self.sanitize_vfpath(vfid, relpath)
        return self.fsop_model.scan_tree(target_path, recursive=recursive)

    async def mkdir(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        *,
        parents: bool = False,
        exist_ok: bool = False,
    ) -> None:
        target_path = self.sanitize_vfpath(vfid, relpath)
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            lambda: target_path.mkdir(0o755, parents=parents, exist_ok=exist_ok),
        )

    async def rmdir(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        *,
        recursive: bool = False,
    ) -> None:
        target_path = self.sanitize_vfpath(vfid, relpath)
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, target_path.rmdir)

    async def move_file(
        self,
        vfid: VFolderID,
        src: PurePosixPath,
        dst: PurePosixPath,
    ) -> None:
        src_path = self.sanitize_vfpath(vfid, src)
        dst_path = self.sanitize_vfpath(vfid, dst)
        await self.fsop_model.move_tree(src_path, dst_path)

    async def move_tree(
        self,
        vfid: VFolderID,
        src: PurePosixPath,
        dst: PurePosixPath,
    ) -> None:
        src_path = self.sanitize_vfpath(vfid, src)
        if not src_path.is_dir():
            raise InvalidAPIParameters(
                msg=f"source path {str(src_path)} is not a directory",
            )
        dst_path = self.sanitize_vfpath(vfid, dst)
        await self.fsop_model.move_tree(src_path, dst_path)

    async def copy_file(
        self,
        vfid: VFolderID,
        src: PurePosixPath,
        dst: PurePosixPath,
    ) -> None:
        src_path = self.sanitize_vfpath(vfid, src)
        if not src_path.is_file():
            raise InvalidAPIParameters(msg=f"source path {str(src_path)} is not a file")
        dst_path = self.sanitize_vfpath(vfid, dst)
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            lambda: dst_path.parent.mkdir(parents=True, exist_ok=True),
        )
        await self.fsop_model.copy_tree(src_path, dst_path)

    async def prepare_upload(self, vfid: VFolderID) -> str:
        vfpath = self.mangle_vfpath(vfid)
        session_id = secrets.token_hex(16)

        def _create_target():
            upload_base_path = vfpath / ".upload"
            upload_base_path.mkdir(exist_ok=True)
            upload_target_path = upload_base_path / session_id
            upload_target_path.touch()

        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, _create_target)
        return session_id

    async def add_file(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        payload: AsyncIterator[bytes],
    ) -> None:
        target_path = self.sanitize_vfpath(vfid, relpath)
        q: janus.Queue[bytes] = janus.Queue()

        def _write(q: janus._SyncQueueProxy[bytes]) -> None:
            with open(target_path, "wb") as f:
                while True:
                    buf = q.get()
                    try:
                        if not buf:
                            return
                        f.write(buf)
                    finally:
                        q.task_done()

        loop = asyncio.get_running_loop()
        write_task: asyncio.Task = asyncio.create_task(
            loop.run_in_executor(None, _write, q.sync_q),  # type: ignore
        )
        try:
            async for buf in payload:
                await q.async_q.put(buf)
            await q.async_q.put(b"")
            await q.async_q.join()
        finally:
            await write_task

    def read_file(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        *,
        chunk_size: int = 0,
    ) -> AsyncIterator[bytes]:
        target_path = self.sanitize_vfpath(vfid, relpath)
        q: janus.Queue[Union[bytes, Exception]] = janus.Queue()
        loop = asyncio.get_running_loop()

        def _read(
            q: janus._SyncQueueProxy[Union[bytes, Exception]],
            chunk_size: int,
        ) -> None:
            try:
                with open(target_path, "rb") as f:
                    while True:
                        buf = f.read(chunk_size)
                        if not buf:
                            return
                        q.put(buf)
            except Exception as e:
                q.put(e)
            finally:
                q.put(b"")

        async def _aiter() -> AsyncIterator[bytes]:
            nonlocal chunk_size
            if chunk_size == 0:
                # get the preferred io block size
                _vfs_stat = await loop.run_in_executor(
                    None,
                    os.statvfs,
                    self.mount_path,
                )
                chunk_size = _vfs_stat.f_bsize
            read_fut = loop.run_in_executor(None, _read, q.sync_q, chunk_size)
            await asyncio.sleep(0)
            try:
                while True:
                    buf = await q.async_q.get()
                    if isinstance(buf, Exception):
                        raise buf
                    yield buf
                    q.async_q.task_done()
                    if not buf:
                        return
            finally:
                await read_fut

        return _aiter()

    async def delete_files(
        self,
        vfid: VFolderID,
        relpaths: Sequence[PurePosixPath],
        *,
        recursive: bool = False,
    ) -> None:
        target_paths = [self.sanitize_vfpath(vfid, p) for p in relpaths]
        for p in target_paths:
            if p.is_dir() and recursive:
                await self.fsop_model.delete_tree(p)
            else:
                await aiofiles.os.remove(p)



import json
import logging
from pathlib import Path
from typing import Any, FrozenSet, Mapping, Optional

from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events import EventDispatcher, EventProducer
from ai.backend.common.types import BinarySize, HardwareMetadata, QuotaScopeID
from ai.backend.logging import BraceStyleAdapter

from ..abc import (
    CAP_FAST_FS_SIZE,
    CAP_METRIC,
    CAP_QUOTA,
    CAP_VFOLDER,
    AbstractFSOpModel,
    AbstractQuotaModel,
    QuotaConfig,
    QuotaUsage,
)
from ..types import CapacityUsage, FSPerfMetric
from ..vfs import BaseFSOpModel, BaseQuotaModel, BaseVolume
from .exceptions import GPFSNoMetricError
from .gpfs_client import GPFSAPIClient

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class GPFSQuotaModel(BaseQuotaModel):
    def __init__(
        self,
        mount_path: Path,
        api_client: GPFSAPIClient,
        fs: str,
        gpfs_owner: str,
    ) -> None:
        super().__init__(mount_path)
        self.api_client = api_client
        self.fs = fs
        self.gpfs_owner = gpfs_owner

    async def create_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        options: Optional[QuotaConfig] = None,
        extra_args: Optional[dict[str, Any]] = None,
    ) -> None:
        qspath = self.mangle_qspath(quota_scope_id)
        await self.api_client.create_fileset(
            self.fs,
            quota_scope_id.pathname,
            path=qspath,
            owner=self.gpfs_owner,
        )
        if options is not None:
            await self.update_quota_scope(quota_scope_id, options)

    async def update_quota_scope(self, quota_scope_id: QuotaScopeID, config: QuotaConfig) -> None:
        await self.api_client.set_quota(self.fs, quota_scope_id.pathname, config.limit_bytes)

    async def describe_quota_scope(self, quota_scope_id: QuotaScopeID) -> Optional[QuotaUsage]:
        if not self.mangle_qspath(quota_scope_id).exists():
            return None

        quotas = await self.api_client.list_fileset_quotas(self.fs, quota_scope_id.pathname)
        custom_defined_quotas = [q for q in quotas if not q.isDefaultQuota]
        if len(custom_defined_quotas) == 0:
            return QuotaUsage(-1, -1)
        quota_info = custom_defined_quotas[0]
        # The units are kilobytes (ref: )
        return QuotaUsage(
            used_bytes=quota_info.blockUsage * 1024 if quota_info.blockUsage is not None else -1,
            limit_bytes=quota_info.blockLimit * 1024 if quota_info.blockLimit is not None else -1,
        )

    async def unset_quota(self, quota_scope_id: QuotaScopeID) -> None:
        await self.api_client.remove_quota(self.fs, quota_scope_id.pathname)

    async def delete_quota_scope(self, quota_scope_id: QuotaScopeID) -> None:
        await self.api_client.remove_fileset(self.fs, quota_scope_id.pathname)


class GPFSOpModel(BaseFSOpModel):
    def __init__(
        self,
        mount_path: Path,
        scandir_limit: int,
        api_client: GPFSAPIClient,
        fs: str,
    ) -> None:
        super().__init__(mount_path, scandir_limit)
        self.api_client = api_client
        self.fs = fs

    async def copy_tree(
        self,
        src_path: Path,
        dst_path: Path,
    ) -> None:
        await self.api_client.copy_folder(
            self.fs,
            src_path,
            self.fs,
            dst_path,
        )


class GPFSVolume(BaseVolume):
    name = "gpfs"
    api_client: GPFSAPIClient

    fs: str

    def __init__(
        self,
        local_config: Mapping[str, Any],
        mount_path: Path,
        *,
        etcd: AsyncEtcd,
        event_dispatcher: EventDispatcher,
        event_producer: EventProducer,
        options: Optional[Mapping[str, Any]] = None,
    ) -> None:
        super().__init__(
            local_config,
            mount_path,
            etcd=etcd,
            options=options,
            event_dispatcher=event_dispatcher,
            event_producer=event_producer,
        )
        verify_ssl = self.config.get("gpfs_verify_ssl", False)
        self.api_client = GPFSAPIClient(
            self.config["gpfs_endpoint"],
            self.config["gpfs_username"],
            self.config["gpfs_password"],
            # follows ssl parameter on https://docs.aiohttp.org/en/v3.10.5/client_reference.html
            ssl=verify_ssl,
        )
        self.fs = self.config["gpfs_fs_name"]
        self.gpfs_owner = self.config.get("gpfs_owner", "1000:1000")

    async def init(self) -> None:
        await super().init()

    async def create_quota_model(self) -> AbstractQuotaModel:
        return GPFSQuotaModel(
            self.mount_path,
            self.api_client,
            self.fs,
            self.gpfs_owner,
        )

    async def create_fsop_model(self) -> AbstractFSOpModel:
        return GPFSOpModel(
            self.mount_path,
            self.local_config["storage-proxy"]["scandir-limit"],
            self.api_client,
            self.fs,
        )

    async def get_capabilities(self) -> FrozenSet[str]:
        return frozenset([CAP_FAST_FS_SIZE, CAP_VFOLDER, CAP_QUOTA, CAP_METRIC])

    async def get_hwinfo(self) -> HardwareMetadata:
        nodes = await self.api_client.list_nodes()
        invalid_status = ["FAILED", "DEGRADED", "ERROR"]
        health_status = "HEALTHY"

        for node in nodes:
            if health_status == "ERROR":
                break
            node_health_statuses = await self.api_client.get_node_health(node)
            for status in node_health_statuses:
                if status.state in invalid_status:
                    health_status = status

        match health_status:
            case "HEALTHY":
                status = "healthy"
            case "DEGRADED":
                status = "degraded"
            case _:
                status = "unavailable"
        cluster_info = await self.api_client.get_cluster_info()
        quotas = await self.api_client.list_quotas(self.fs)
        return {
            "status": status,
            "status_info": None,
            "metadata": {
                "quota": json.dumps([q.to_json() for q in quotas]),
                "cluster_info": json.dumps(cluster_info),
            },
        }

    async def get_fs_usage(self) -> CapacityUsage:
        storage_pools = await self.api_client.list_fs_pools(self.fs)
        free, total = 0, 0
        for _pool in storage_pools:
            pool = await self.api_client.get_fs_pool(self.fs, _pool.storagePoolName)
            if pool.totalDataInKB is None or pool.freeDataInKB is None:
                continue
            total += pool.totalDataInKB
            free += pool.freeDataInKB
        return CapacityUsage(
            used_bytes=BinarySize(total - free) * 1024,
            capacity_bytes=BinarySize(total) * 1024,
        )

    async def get_performance_metric(self) -> FSPerfMetric:
        # ref: https://www.ibm.com/docs/en/spectrum-scale/5.0.3?topic=2-perfmondata-get
        query = (
            "metrics "
            "avg(gpfs_ns_read_ops),"
            "avg(gpfs_ns_write_ops),"
            "avg(gpfs_ns_bytes_read),"
            "avg(gpfs_ns_bytes_written),"
            "avg(gpfs_ns_max_disk_wait_rd),"
            "avg(gpfs_ns_max_disk_wait_wr) "
            "last 1 "
            "bucket_size 10"
        )
        try:
            metrics = await self.api_client.get_metric(query)
            latest_metric = metrics["performanceData"]["rows"][-1]["values"]
        except (KeyError, IndexError):
            raise GPFSNoMetricError
        return FSPerfMetric(
            iops_read=latest_metric[0] or 0,
            iops_write=latest_metric[1] or 0,
            io_bytes_read=latest_metric[2] or 0,
            io_bytes_write=latest_metric[3] or 0,
            io_usec_read=latest_metric[4] or 0,
            io_usec_write=latest_metric[5] or 0,
        )



from typing import Optional


class VASTError(Exception):
    pass


class VASTAPIError(VASTError):
    message: str

    def __init__(self, message: Optional[str] = None, *args):
        super().__init__(message, *args)

        self.message = message or ""

    def __str__(self) -> str:
        return type(self).__name__


class VASTUnauthorizedError(VASTAPIError):
    pass


class VASTInvalidParameterError(VASTAPIError):
    pass


class VASTNotFoundError(VASTAPIError):
    pass


class VASTClusterNotFoundError(VASTAPIError):
    pass


class VASTUnknownError(VASTAPIError):
    pass



from typing import Optional


class WekaError(Exception):
    pass


class WekaInitError(WekaError):
    pass


class WekaAPIError(WekaError):
    message: str

    def __init__(self, message: Optional[str] = None, *args):
        super().__init__(message, *args)

        self.message = message or ""

    def __str__(self) -> str:
        return type(self).__name__


class WekaInvalidBodyError(WekaAPIError):
    pass


class WekaUnauthorizedError(WekaAPIError):
    pass


class WekaNotFoundError(WekaAPIError):
    pass


class WekaInternalError(WekaAPIError):
    pass


class WekaNoMetricError(WekaError):
    pass
